/// Copyright (c) 2022 Razeware LLC
///
/// Permission is hereby granted, free of charge, to any person obtaining a copy
/// of this software and associated documentation files (the "Software"), to deal
/// in the Software without restriction, including without limitation the rights
/// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
/// copies of the Software, and to permit persons to whom the Software is
/// furnished to do so, subject to the following conditions:
///
/// The above copyright notice and this permission notice shall be included in
/// all copies or substantial portions of the Software.
///
/// Notwithstanding the foregoing, you may not use, copy, modify, merge, publish,
/// distribute, sublicense, create a derivative work, and/or sell copies of the
/// Software in any work that is designed, intended, or marketed for pedagogical or
/// instructional purposes related to programming, coding, application development,
/// or information technology.  Permission for such use, copying, modification,
/// merger, publication, distribution, sublicensing, creation of derivative works,
/// or sale is expressly withheld.
///
/// This project and source code may use libraries or frameworks that are
/// released under various Open-Source licenses. Use of those libraries and
/// frameworks are governed by their own individual licenses.
///
/// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
/// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
/// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
/// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
/// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
/// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
/// THE SOFTWARE.

import MetalKit

// swiftlint:disable implicitly_unwrapped_optional

class Renderer: NSObject {
  static var device: MTLDevice!
  static var commandQueue: MTLCommandQueue!
  static var library: MTLLibrary!
  var options: Options

  var pipelineState: MTLRenderPipelineState!
  let depthStencilState: MTLDepthStencilState?

    var lastTime: Double = CFAbsoluteTimeGetCurrent()

  var uniforms = Uniforms()
  var params = Params()
    
    var quad: Quad

    
    lazy var scene = GameScene()

  // the models to render


  init(metalView: MTKView, options: Options) {
    guard
      let device = MTLCreateSystemDefaultDevice(),
      let commandQueue = device.makeCommandQueue() else {
        fatalError("GPU not available")
    }
    Renderer.device = device
    Renderer.commandQueue = commandQueue
    metalView.device = device
      
      quad = Quad(device: Renderer.device, scale: 0.8)
        

    // create the shader function library
    let library = device.makeDefaultLibrary()
    Self.library = library
    let vertexFunction = library?.makeFunction(name: "vertex_main")
    let fragmentFunction =
      library?.makeFunction(name: "fragment_main")

    // create the pipeline state
    let pipelineDescriptor = MTLRenderPipelineDescriptor()
    pipelineDescriptor.vertexFunction = vertexFunction
    pipelineDescriptor.fragmentFunction = fragmentFunction
    pipelineDescriptor.colorAttachments[0].pixelFormat =
      metalView.colorPixelFormat
    pipelineDescriptor.depthAttachmentPixelFormat = .depth32Float
    do {
//      pipelineDescriptor.vertexDescriptor =
//        MTLVertexDescriptor.defaultLayout
      pipelineDescriptor.vertexDescriptor = MTKMetalVertexDescriptorFromModelIO(.quadLayout)
      pipelineState =
        try device.makeRenderPipelineState(
          descriptor: pipelineDescriptor)
    } catch let error {
      fatalError(error.localizedDescription)
    }
    self.options = options
    depthStencilState = Renderer.buildDepthStencilState()
    super.init()
    metalView.clearColor = MTLClearColor(
      red: 0.93,
      green: 0.97,
      blue: 1.0,
      alpha: 1.0)
    metalView.depthStencilPixelFormat = .depth32Float
    metalView.delegate = self
    mtkView(metalView, drawableSizeWillChange: metalView.bounds.size)
  }

  static func buildDepthStencilState() -> MTLDepthStencilState? {
    let descriptor = MTLDepthStencilDescriptor()
    descriptor.depthCompareFunction = .less
    descriptor.isDepthWriteEnabled = true
    return Renderer.device.makeDepthStencilState(
      descriptor: descriptor)
  }
}

extension Renderer: MTKViewDelegate {
  func mtkView(
    _ view: MTKView,
    drawableSizeWillChange size: CGSize
  ) {
      scene.update(size: size)
  }

  func draw(in view: MTKView) {
    guard
      let commandBuffer = Renderer.commandQueue.makeCommandBuffer(),
      let descriptor = view.currentRenderPassDescriptor,
      let renderEncoder =
        commandBuffer.makeRenderCommandEncoder(
          descriptor: descriptor) else {
        return
    }

    
    renderEncoder.setDepthStencilState(depthStencilState)
    renderEncoder.setRenderPipelineState(pipelineState)

    // update and render
      let currentTime = CFAbsoluteTimeGetCurrent()
      let deltaTime = Float(currentTime - lastTime)
      lastTime = currentTime
      scene.update(deltaTime: deltaTime)

      
      uniforms.viewMatrix = scene.camera.viewMatrix
      uniforms.projectionMatrix = scene.camera.projectionMatrix

      
//      for model in scene.models {
//        model.render(
//          encoder: renderEncoder,
//          uniforms: uniforms,
//          params: params)
//      }
      var tiling: UInt32 = 1
      params.tiling = tiling
      
      var transform = Transform()

      uniforms.modelMatrix = transform.modelMatrix

      renderEncoder.setVertexBytes(
        &uniforms,
        length: MemoryLayout<Uniforms>.stride,
        index: UniformsBuffer.index)
      
      //Swift.print("uniforms ", UniformsBuffer.index)  11

      renderEncoder.setFragmentBytes(
        &params,
        length: MemoryLayout<Uniforms>.stride,
        index: ParamsBuffer.index)
      
      //Swift.print("params ", ParamsBuffer.index)   12
      
      renderEncoder.setVertexBuffer(
        quad.vertexBuffer,
        offset: 0,
        index: 0)
      
//      renderEncoder.setVertexBuffer(
//        quad.indexBuffer,
//        offset: 0,
//        index: 1)


//      renderEncoder.drawPrimitives(
//        type: .triangle,
//        vertexStart: 0,
//        vertexCount: quad.indices.count)
      
      renderEncoder.drawIndexedPrimitives(
        type: .triangle,
        indexCount: quad.indices.count,
        indexType: .uint16,
        indexBuffer: quad.indexBuffer,
        indexBufferOffset: 0)



      
    // end update and render

    renderEncoder.endEncoding()
    guard let drawable = view.currentDrawable else {
      return
    }
    commandBuffer.present(drawable)
    commandBuffer.commit()
  }
}
